"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Database_level, _Database_rules;
Object.defineProperty(exports, "__esModule", { value: true });
exports.Database = void 0;
const storage_1 = require("../storage");
const lock_1 = require("./lock");
const query_1 = require("./query");
const logging_1 = require("@hibas123/logging");
const nanoid = require("nanoid");
const utils_1 = require("@hibas123/utils");
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
const longNanoId = nanoid.customAlphabet(ALPHABET, 32);
const shortNanoId = nanoid.customAlphabet(ALPHABET, 16);
const DummyRuleEngine = {
    hasPermission: () => true,
};
class Database {
    constructor(database_path, ruleEngine) {
        this.database_path = database_path;
        this.logging = new logging_1.LoggingBase({ console: false });
        _Database_level.set(this, storage_1.getRawLevelDB(this.database_path));
        _Database_rules.set(this, void 0);
        this.connectionCount = 0;
        this.locks = new lock_1.default();
        this.collectionLocks = new lock_1.default();
        this.changeListener = new Map();
        this.collectionChangeListener = new utils_1.Observable();
        if (ruleEngine)
            __classPrivateFieldSet(this, _Database_rules, ruleEngine, "f");
        else
            __classPrivateFieldSet(this, _Database_rules, DummyRuleEngine, "f");
    }
    static getKey(collectionid, documentid) {
        return `${collectionid || ""}/${documentid || ""}`;
    }
    get data() {
        return __classPrivateFieldGet(this, _Database_level, "f").data;
    }
    get collections() {
        return __classPrivateFieldGet(this, _Database_level, "f").collection;
    }
    get rules() {
        return __classPrivateFieldGet(this, _Database_rules, "f");
    }
    set rules(engine) {
        if (!engine) {
            __classPrivateFieldSet(this, _Database_rules, DummyRuleEngine, "f");
        }
        else {
            __classPrivateFieldSet(this, _Database_rules, engine, "f");
        }
    }
    toJSON() {
        return {
            database_path: this.database_path,
            rules: __classPrivateFieldGet(this, _Database_rules, "f"),
            connections: this.connectionCount,
        };
    }
    async resolve(path, create = false) {
        path = [...path]; // Create modifiable copy
        let collectionID = undefined;
        let documentKey = path.length % 2 === 0 ? path.pop() : undefined;
        let key = path.join("/");
        const lock = await this.collectionLocks.lock(key);
        try {
            collectionID = await this.collections
                .get(key)
                .then((r) => r.toString())
                .catch(storage_1.resNull);
            if (!collectionID && create) {
                collectionID = longNanoId();
                await this.collections.put(key, collectionID);
                setImmediate(() => {
                    this.collectionChangeListener.send({
                        id: collectionID,
                        key,
                        type: "create",
                    });
                });
            }
        }
        finally {
            lock();
        }
        return {
            collection: collectionID,
            document: documentKey,
            collectionKey: key,
        };
    }
    sendChanges(changes) {
        let col = new Map();
        changes.forEach((change) => {
            let e = col.get(change.collection);
            if (!e) {
                e = new Map();
                col.set(change.collection, e);
            }
            let d = e.get(change.document);
            if (!d) {
                d = [];
                e.set(change.document, d);
            }
            d.push(change);
        });
        setImmediate(() => {
            for (let [collection, documents] of col.entries()) {
                let collectionChanges = [];
                for (let [document, documentChanges] of documents.entries()) {
                    let s = this.changeListener.get(Database.getKey(collection, document));
                    if (s)
                        s.forEach((e) => setImmediate(() => e(documentChanges)));
                    collectionChanges.push(...documentChanges);
                }
                let s = this.changeListener.get(Database.getKey(collection));
                if (s)
                    s.forEach((e) => setImmediate(() => e(collectionChanges)));
            }
        });
    }
    validate(query) {
        const inv = new query_1.QueryError("Malformed query!");
        if (!query || typeof query !== "object")
            throw inv;
        if (!query.type)
            throw inv;
        if (!query.path)
            throw inv;
    }
    async run(queries, session) {
        let resolve = [];
        const addToResolve = (path, create) => {
            let entry = resolve.find((e) => {
                //TODO: Find may be slow...
                if (e.path.length !== path.length)
                    return false;
                for (let i = 0; i < e.path.length; i++) {
                    if (e.path[i] !== path[i])
                        return false;
                }
                return true;
            });
            if (!entry) {
                entry = {
                    path,
                    create,
                };
                resolve.push(entry);
            }
            entry.create = entry.create || create;
            return entry;
        };
        const isBatch = queries.length > 1;
        let parsed = queries.map((rawQuery) => {
            this.logging.debug("Running query:", rawQuery.type);
            this.validate(rawQuery);
            const isCollection = rawQuery.path.length % 2 === 1;
            let query = isCollection
                ? new query_1.CollectionQuery(this, session, rawQuery)
                : new query_1.DocumentQuery(this, session, rawQuery);
            if (isBatch && !query.batchCompatible)
                throw new Error("There are queries that are not batch compatible!");
            let path = addToResolve(rawQuery.path, query.createCollection);
            if (query.additionalLock)
                addToResolve(query.additionalLock);
            return {
                path,
                query,
            };
        });
        resolve = resolve.sort((a, b) => a.path.length - b.path.length);
        let locks = [];
        for (let e of resolve) {
            let { collection, document, collectionKey } = await this.resolve(e.path, e.create);
            e.resolved = [collection, document, collectionKey];
            locks.push(await this.locks.lock(collection, document));
        }
        let result = [];
        try {
            let batch = this.data.batch();
            let changes = [];
            for (let e of parsed) {
                result.push(await e.query.run(e.path.resolved[0], e.path.resolved[1], batch, e.path.resolved[2]));
                changes.push(...e.query.changes);
            }
            if (batch.length > 0)
                await batch.write();
            this.sendChanges(changes);
        }
        finally {
            locks.forEach((lock) => lock());
        }
        if (isBatch)
            return result;
        else
            return result[0];
    }
    async snapshot(rawQuery, session, onchange) {
        this.logging.debug("Snaphot request:", rawQuery.path);
        this.validate(rawQuery);
        if (rawQuery.type !== "snapshot")
            throw new Error("Invalid query type!");
        const isCollection = rawQuery.path.length % 2 === 1;
        let query = isCollection
            ? new query_1.CollectionQuery(this, session, rawQuery, true)
            : new query_1.DocumentQuery(this, session, rawQuery, true);
        const { unsubscribe, value } = await query.snapshot(onchange);
        const id = shortNanoId();
        session.subscriptions.set(id, unsubscribe);
        return {
            id,
            snaphot: value,
        };
    }
    async unsubscribe(id, session) {
        let query = session.subscriptions.get(id);
        if (query) {
            query();
            session.subscriptions.delete(id);
        }
    }
    async stop() {
        await this.data.close();
    }
    async delete() {
        await this.stop();
        return storage_1.deleteRawLevelDB(this.database_path);
    }
    async runCleanup() {
        const should = await new Promise((yes, no) => {
            const stream = this.collections.iterator({
                keyAsBuffer: false,
                valueAsBuffer: false,
            });
            const collections = new Set();
            const onValue = (err, key, value) => {
                if (err) {
                    this.logging.error(err);
                    stream.end((err) => this.logging.error(err));
                    no(err);
                }
                if (!key && !value) {
                    yes(collections);
                }
                else {
                    collections.add(value);
                    stream.next(onValue);
                }
            };
            stream.next(onValue);
        });
        const existing = await new Promise((yes, no) => {
            const stream = this.data.iterator({
                keyAsBuffer: false,
                values: false,
            });
            const collections = new Set();
            const onValue = (err, key, value) => {
                if (err) {
                    this.logging.error(err);
                    stream.end((err) => this.logging.error(err));
                    no(err);
                }
                if (!key && !value) {
                    yes(collections);
                }
                else {
                    let coll = key.split("/")[0];
                    collections.add(coll);
                    stream.next(onValue);
                }
            };
            stream.next(onValue);
        });
        const toDelete = new Set();
        existing.forEach((collection) => {
            if (!should.has(collection))
                toDelete.add(collection);
        });
        for (let collection of toDelete) {
            const batch = this.data.batch();
            let gt = Buffer.from(collection + "/ ");
            gt[gt.length - 1] = 0;
            let lt = Buffer.alloc(gt.length);
            lt.set(gt);
            lt[gt.length - 1] = 0xff;
            await new Promise((yes, no) => {
                const stream = this.data.iterator({
                    keyAsBuffer: false,
                    values: false,
                    gt,
                    lt,
                });
                const onValue = (err, key, value) => {
                    if (err) {
                        this.logging.error(err);
                        stream.end((err) => this.logging.error(err));
                        no(err);
                    }
                    if (!key && !value) {
                        yes();
                    }
                    else {
                        batch.del(key);
                        stream.next(onValue);
                    }
                };
                stream.next(onValue);
            });
            await batch.write();
        }
        return Array.from(toDelete.values());
    }
}
exports.Database = Database;
_Database_level = new WeakMap(), _Database_rules = new WeakMap();
//# sourceMappingURL=database.js.map