import { Database, Change, ChangeTypes } from "./database";
import * as MSGPack from "msgpack5";
import Session from "./session";
import { LevelUpChain } from "levelup";
import { Operations } from "../rules";
export declare type IWriteQueries = "set" | "update" | "delete" | "add";
export declare type ICollectionQueries = "get" | "add" | "keys" | "delete-collection" | "list";
export declare type IDocumentQueries = "get" | "set" | "update" | "delete";
export interface ITypedQuery<T> {
    path: string[];
    type: T;
    data?: any;
    options?: any;
}
export declare type IQuery = ITypedQuery<ICollectionQueries | IDocumentQueries | "snapshot">;
export declare const MP: MSGPack.MessagePack;
declare type Runner = (collection: string, document: string, batch: LevelUpChain, collectionKey: string) => any;
interface IPreparedQuery {
    createCollection: boolean;
    needDocument: boolean;
    batchCompatible: boolean;
    runner: Runner;
    permission: Operations;
    additionalLock?: string[];
}
interface DocRes {
    id: string;
    data: any;
}
export declare abstract class Query {
    protected database: Database;
    protected session: Session;
    protected query: IQuery;
    /**
     * Returns true if the path only contains valid characters and false if it doesn't
     * @param path Path to be checked
     */
    private validatePath;
    changes: Change[];
    readonly createCollection: boolean;
    readonly needDocument: boolean;
    readonly batchCompatible: boolean;
    readonly additionalLock?: string[];
    readonly permission: Operations;
    private readonly _runner;
    constructor(database: Database, session: Session, query: IQuery, snapshot?: boolean);
    protected abstract prepare(query: IQuery): IPreparedQuery;
    protected getDoc(collection: string, document: string): Promise<any>;
    protected sendChange(collection: string, document: string, type: ChangeTypes, data: any): void;
    protected static getConstructorParams(query: Query): [Database, Session, IQuery];
    protected abstract checkChange(change: Change): boolean;
    protected abstract firstSend(collection: string, document: string): Promise<any>;
    run(collection: string, document: string, batch: LevelUpChain, collectionKey: string): any;
    snapshot(onChange: (change: (DocRes & {
        type: ChangeTypes;
    })[]) => void): Promise<{
        unsubscribe: () => void;
        value: any;
    }>;
}
export declare class DocumentQuery extends Query {
    prepare(query: IQuery): IPreparedQuery;
    private get;
    private set;
    private update;
    private delete;
    checkChange(change: Change): boolean;
    firstSend(collection: string, document: string): Promise<any>;
    static fromQuery(query: Query): DocumentQuery;
}
declare type FieldPath = string;
declare type WhereFilterOp = "<" | "<=" | "==" | ">=" | ">" | "array-contains" | "in" | "array-contains-any";
interface IQueryWhereVerbose {
    fieldPath: FieldPath;
    opStr: WhereFilterOp;
    value: any;
}
declare type IQueryWhereArray = [FieldPath, WhereFilterOp, any];
declare type IQueryWhere = IQueryWhereArray | IQueryWhereVerbose;
export declare class CollectionQuery extends Query {
    private _addId;
    prepare(query: any): IPreparedQuery;
    private _where;
    set where(value: IQueryWhere[]);
    limit: number;
    add(collection: string, document: string, batch: LevelUpChain, collectionKey: string): Promise<string>;
    private getStreamOptions;
    keys(collection: string): Promise<string[]>;
    private _getFieldValue;
    private _fitsWhere;
    get(collection: string): Promise<DocRes[]>;
    checkChange(change: Change): boolean;
    firstSend(collection: string): Promise<DocRes[]>;
    collections(): Promise<string[]>;
    deleteCollection(collection: string, document: string, _b: LevelUpChain, collectionKey: string): Promise<void>;
    static fromQuery(query: Query): CollectionQuery;
}
export declare class QueryError extends Error {
    constructor(message: string);
}
export {};
