"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.QueryError = exports.CollectionQuery = exports.DocumentQuery = exports.Query = exports.MP = void 0;
const database_1 = require("./database");
const storage_1 = require("../storage");
const nanoid = require("nanoid");
const MSGPack = require("msgpack5");
exports.MP = MSGPack({});
// MSGPack.initialize(2 ** 20);
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
const longNanoID = nanoid.customAlphabet(ALPHABET, 32);
const { encode, decode } = exports.MP;
class Query {
    constructor(database, session, query, snapshot = false) {
        this.database = database;
        this.session = session;
        this.query = query;
        this.changes = [];
        if (query.path.length > 10) {
            throw new QueryError("Path is to long. Path is only allowed to be 10 Layers deep!");
        }
        if (!this.validatePath(query.path)) {
            throw new QueryError("Path can only contain a-z A-Z 0-9  '-' '-' '<' and '>' ");
        }
        if (!snapshot) {
            let data = this.prepare(query);
            this.createCollection = data.createCollection;
            this.needDocument = data.needDocument;
            this.batchCompatible = data.batchCompatible;
            this.additionalLock = data.additionalLock;
            this.permission = data.permission;
            this._runner = data.runner;
        }
    }
    /**
     * Returns true if the path only contains valid characters and false if it doesn't
     * @param path Path to be checked
     */
    validatePath(path) {
        return path.every((e) => (e.match(/[^a-zA-Z0-9_\-\<\>]/g) || []).length === 0);
    }
    getDoc(collection, document) {
        return this.database.data
            .get(database_1.Database.getKey(collection, document), { asBuffer: true })
            .then((res) => decode(res))
            .catch(storage_1.resNull);
    }
    sendChange(collection, document, type, data) {
        let change = {
            type,
            document,
            collection,
            data,
            sender: this.session.id,
        };
        this.database.logging.debug("Sending change:", change);
        this.changes.push(change);
    }
    static getConstructorParams(query) {
        return [query.database, query.session, query.query];
    }
    run(collection, document, batch, collectionKey) {
        let perm = this.database.rules.hasPermission(this.query.path, this.permission, this.session);
        if (!perm)
            throw new QueryError("No permission!");
        // this.query.path = perm.path;
        return this._runner.call(this, collection, document, batch, collectionKey);
    }
    async snapshot(onChange) {
        let perm = this.database.rules.hasPermission(this.query.path, "read", this.session);
        if (!perm) {
            throw new QueryError("No permission!");
        }
        const receivedChanges = (changes) => {
            let res = changes
                .filter((change) => this.checkChange(change))
                .map((change) => {
                return {
                    id: change.document,
                    data: change.data,
                    type: change.type,
                };
            });
            if (res.length > 0)
                onChange(res);
        };
        const unsub = this.database.collectionChangeListener.subscribe((change) => {
            if (change.key === collectionKey) {
                if (change.type === "create")
                    addSubscriber(change.id);
                else
                    removeSubscriber(); // Send delete for all elements (Don't know how to do this...)
            }
        });
        let { collection, document, collectionKey } = await this.database.resolve(this.query.path);
        let oldKey = undefined;
        const removeSubscriber = () => {
            if (!oldKey)
                return;
            let s = this.database.changeListener.get(oldKey);
            if (s) {
                s.delete(receivedChanges);
                if (s.size <= 0)
                    this.database.changeListener.delete(oldKey);
            }
            oldKey = undefined;
        };
        const addSubscriber = (collection) => {
            let key = database_1.Database.getKey(collection, document);
            if (oldKey !== key) {
                if (oldKey !== undefined)
                    removeSubscriber();
                let s = this.database.changeListener.get(key);
                if (!s) {
                    s = new Set();
                    this.database.changeListener.set(key, s);
                }
                s.add(receivedChanges);
            }
        };
        if (collection) {
            addSubscriber(collection);
        }
        return {
            unsubscribe: () => {
                unsub();
                removeSubscriber();
            },
            value: await this.firstSend(collection, document),
        };
    }
}
exports.Query = Query;
class DocumentQuery extends Query {
    prepare(query) {
        let type = query.type;
        switch (type) {
            case "get":
                return {
                    batchCompatible: false,
                    createCollection: false,
                    needDocument: false,
                    permission: "read",
                    runner: this.get,
                };
            case "set":
                return {
                    batchCompatible: true,
                    createCollection: true,
                    needDocument: true,
                    permission: "write",
                    runner: this.set,
                };
            case "update":
                return {
                    batchCompatible: true,
                    createCollection: true,
                    needDocument: true,
                    permission: "write",
                    runner: this.update,
                };
            case "delete":
                return {
                    batchCompatible: true,
                    createCollection: false,
                    needDocument: true,
                    permission: "write",
                    runner: this.delete,
                };
            default:
                throw new Error("Invalid query type: " + type);
        }
    }
    async get(collection, document) {
        if (!collection || !document) {
            return null;
        }
        return this.getDoc(collection, document);
    }
    async set(collection, document, batch) {
        const { data, options } = this.query;
        if (data === null)
            return this.delete(collection, document, batch);
        let isNew = !(await this.getDoc(collection, document));
        batch.put(database_1.Database.getKey(collection, document), encode(data));
        this.sendChange(collection, document, isNew ? "added" : "modified", data);
    }
    async update(collection, document, batch) {
        const updateData = this.query.data;
        let data = await this.getDoc(collection, document);
        let isNew = false;
        if (!data) {
            isNew = true;
            data = {};
        }
        for (let path in updateData) {
            const toUpdate = updateData[path];
            let d = data;
            let parts = path.split(".");
            while (parts.length > 1) {
                let seg = parts.shift();
                if (!data[seg])
                    data[seg] = {};
                d = data[seg];
            }
            const last = parts[0];
            switch (toUpdate.type) {
                case "value":
                    d[last] = toUpdate.value;
                    break;
                case "increment":
                    if (d[last] === undefined || d[last] === null)
                        d[last] = toUpdate.value;
                    else if (typeof d[last] !== "number") {
                        throw new QueryError("Field is no number!");
                    }
                    else {
                        d[last] += toUpdate.value;
                    }
                    break;
                case "timestamp":
                    d[last] = new Date().valueOf();
                    break;
                case "push":
                    if (d[last] === undefined || d[last] === null)
                        d[last] = [toUpdate.value];
                    else if (Array.isArray(d[last])) {
                        d[last].push(toUpdate.value);
                    }
                    else {
                        throw new QueryError("Field is not array!");
                    }
                    break;
                default:
                    throw new QueryError("Invalid update type: " + toUpdate.type);
            }
        }
        if (batch) {
            batch.put(database_1.Database.getKey(collection, document), encode(data));
        }
        else {
            await this.database.data.put(database_1.Database.getKey(collection, document), encode(data).slice(0));
        }
        this.sendChange(collection, document, isNew ? "added" : "modified", data);
    }
    async delete(collection, document, batch) {
        if (batch) {
            batch.del(database_1.Database.getKey(collection, document));
        }
        else {
            await this.database.data.del(database_1.Database.getKey(collection, document));
        }
        this.sendChange(collection, document, "deleted", null);
    }
    checkChange(change) {
        return true;
    }
    firstSend(collection, document) {
        return this.get(collection, document);
    }
    static fromQuery(query) {
        return new DocumentQuery(...Query.getConstructorParams(query));
    }
}
exports.DocumentQuery = DocumentQuery;
class CollectionQuery extends Query {
    constructor() {
        super(...arguments);
        this.limit = -1;
    }
    prepare(query) {
        switch (query.type) {
            case "add":
                this._addId = longNanoID();
                return {
                    batchCompatible: true,
                    createCollection: true,
                    needDocument: false,
                    runner: this.add,
                    permission: "write",
                    additionalLock: [...query.path, this._addId],
                };
            case "get":
                const limit = (query.options || {}).limit;
                if (limit)
                    this.limit = limit;
                const where = (query.options || {}).where;
                if (where)
                    this.where = where;
                return {
                    batchCompatible: false,
                    createCollection: false,
                    needDocument: false,
                    permission: "read",
                    runner: this.get,
                };
            case "keys":
                return {
                    batchCompatible: false,
                    createCollection: false,
                    needDocument: false,
                    permission: "list",
                    runner: this.keys,
                };
            case "list":
                return {
                    batchCompatible: false,
                    createCollection: false,
                    needDocument: false,
                    permission: "read",
                    runner: this.keys,
                };
            case "delete-collection":
                return {
                    batchCompatible: false,
                    createCollection: false,
                    needDocument: false,
                    permission: "write",
                    runner: this.deleteCollection,
                };
            // run = () => q.deleteCollection();
            // break;
            default:
                throw new Error("Invalid query!");
        }
    }
    set where(value) {
        const invalidWhere = new QueryError("Invalid Where");
        if (!Array.isArray(value))
            throw invalidWhere;
        let c = [];
        this._where = value.map((cond) => {
            this.database.logging.debug("Query Condition", cond);
            if (Array.isArray(cond)) {
                if (cond.length !== 3)
                    throw invalidWhere;
                return cond;
            }
            else {
                if (cond &&
                    typeof cond === "object" &&
                    "fieldPath" in cond &&
                    "opStr" in cond &&
                    "value" in cond) {
                    return [cond.fieldPath, cond.opStr, cond.value];
                }
                else {
                    throw invalidWhere;
                }
            }
        });
    }
    async add(collection, document, batch, collectionKey) {
        let q = new DocumentQuery(this.database, this.session, {
            type: "set",
            path: this.additionalLock,
            data: this.query.data,
            options: this.query.options,
        });
        await q.run(collection, this._addId, batch, collectionKey);
        this.changes = q.changes;
        return this._addId;
    }
    getStreamOptions(collection) {
        let gt = Buffer.from(database_1.Database.getKey(collection) + " ");
        gt[gt.length - 1] = 0;
        let lt = Buffer.alloc(gt.length);
        lt.set(gt);
        lt[gt.length - 1] = 0xff;
        return {
            gt,
            lt,
        };
    }
    async keys(collection) {
        if (!collection)
            return [];
        return new Promise((yes, no) => {
            let keys = [];
            const stream = this.database.data.createKeyStream({
                ...this.getStreamOptions(collection),
                keyAsBuffer: false,
            });
            stream.on("data", (key) => {
                let s = key.split("/", 2);
                if (s.length > 1)
                    keys.push(s[1]);
            });
            stream.on("end", () => yes(keys));
            stream.on("error", no);
        });
    }
    _getFieldValue(data, path) {
        let parts = path.split(".");
        let d = data;
        while (parts.length > 0) {
            let seg = parts.shift();
            d = data[seg];
            if (d === undefined || d === null)
                break; // Undefined/Null has no other fields!
        }
        return d;
    }
    _fitsWhere(data) {
        if (this._where && this._where.length > 0) {
            return this._where.every(([fieldPath, opStr, value]) => {
                let val = this._getFieldValue(data, fieldPath);
                switch (opStr) {
                    case "<":
                        return val < value;
                    case "<=":
                        return val <= value;
                    case "==":
                        return val == value;
                    case ">=":
                        return val >= value;
                    case ">":
                        return val > value;
                    case "array-contains":
                        if (Array.isArray(val)) {
                            return val.some((e) => e === value);
                        }
                        return false;
                    // case "array-contains-any":
                    case "in":
                        if (typeof val === "object") {
                            return value in val;
                        }
                        return false;
                    default:
                        throw new QueryError("Invalid where operation " + opStr);
                }
            });
        }
        return true;
    }
    async get(collection) {
        if (!collection)
            return [];
        return new Promise((yes, no) => {
            const stream = this.database.data.iterator({
                ...this.getStreamOptions(collection),
                keyAsBuffer: false,
                valueAsBuffer: true,
            });
            let values = [];
            const onValue = (err, key, value) => {
                if (err) {
                    no(err);
                    stream.end((err) => this.database.logging.error(err));
                }
                else {
                    if (!key && !value) {
                        // END
                        this.database.logging.debug("Checked all!");
                        yes(values);
                    }
                    else {
                        let s = key.split("/", 2);
                        if (s.length <= 1)
                            return;
                        const id = s[1];
                        let data = decode(value);
                        if (this._fitsWhere(data)) {
                            if (this.limit < 0 || values.length < this.limit) {
                                values.push({
                                    id,
                                    data,
                                });
                            }
                            else {
                                stream.end((err) => (err ? no(err) : yes(values)));
                                return;
                            }
                        }
                        stream.next(onValue);
                    }
                }
            };
            stream.next(onValue);
        });
    }
    checkChange(change) {
        return this._fitsWhere(change.data);
    }
    firstSend(collection) {
        return this.get(collection);
    }
    async collections() {
        if (!this.session.root)
            throw new QueryError("No Permission!");
        return new Promise((yes, no) => {
            let keys = [];
            const stream = this.database.data.createKeyStream({
                keyAsBuffer: false,
            });
            stream.on("data", (key) => keys.push(key.split("/")));
            stream.on("end", () => yes(keys));
            stream.on("error", no);
        });
    }
    async deleteCollection(collection, document, _b, collectionKey) {
        if (!this.session.root)
            throw new QueryError("No Permission!");
        //TODO: Lock whole collection!
        let batch = this.database.data.batch();
        try {
            if (collection) {
                let documents = await this.keys(collection);
                // Logging.debug("To delete:", documents)
                for (let document of documents) {
                    batch.del(database_1.Database.getKey(collection, document));
                }
                await batch.write();
                batch = undefined;
                await this.database.collections.del(collectionKey);
                this.database.collectionChangeListener.send({
                    id: collection,
                    key: collectionKey,
                    type: "delete",
                });
            }
        }
        finally {
            if (batch)
                batch.clear();
        }
    }
    static fromQuery(query) {
        return new CollectionQuery(...Query.getConstructorParams(query));
    }
}
exports.CollectionQuery = CollectionQuery;
class QueryError extends Error {
    constructor(message) {
        super(message);
    }
}
exports.QueryError = QueryError;
//# sourceMappingURL=query.js.map